#!/usr/bin/perl -w
use strict;
use lib 'lib';
use Javascript::PurePerl;
use Carp qw(carp);

my $JS = q{function f() {return 3;};
var a = 42; var b,c; perl.print(a); var d = 4;
var e = function(g,h) {
  perl.print(g+h);
};
perl.print( f()+d );
e(10,32);
};

my $tree = Javascript::PurePerl->new()->parse( code => $JS );

sub eliminate_nodes {
  my ($nodes) = @_;
  my @nodes = $tree->findnodes($nodes);
  for my $node (@nodes) {
    my $parent = $node->parentNode();
    my $after_node = $node;
    for my $child ($node->childNodes) {
      $node->removeChild($child);
      $after_node = $parent->insertAfter($child, $after_node);
    };
    $parent->removeChild($node);
  };
};

sub prepend {
  my ($node, @nodes) = @_;
  carp "no parent" unless $node;
  for (reverse @nodes) {
    ($_) = textnode( $_ )
      unless ref $_;
    if ($node->firstChild) {
      $node->insertBefore( $_, $node->firstChild );
    } else {
      $node->appendChild( $_ );
    };
  };
  $node;
};

sub append {
  my ($node, @nodes) = @_;
  carp "no parent" unless $node;
  for (@nodes) {
    ($_) = textnode( $_ )
      unless ref $_;
    if ($node->lastChild) {
      $node->insertAfter( $_, $node->lastChild );
    } else {
      $node->appendChild( $_ );
    };
  };
  $node;
};

sub textnode { map { $tree->createTextNode($_) } @_ };

print $tree->toString,"\n";

# translate variable names
for my $node ($tree->findnodes('//Identifier[@name = "Callee"]')) {
  # now check whether it's a variable or a sub name:
  # - a variable name would be within a VariableDeclaration that is contained
  #   in a parent of the current node
  # - but it has to be an older sibling of the relevant ancestor
  # - that VariableDeclaration must have our name as "name" attribute
  # ancestor::*/preceding-sibling::*//VariableDeclaration[@name=name]
  my $name = $node->textContent();  
  my (@same_name_declarations) = $node->findnodes( sprintf q{ancestor::*/preceding-sibling::*//VariableDeclaration/Identifier[text()='%s']}, $name );
  if (@same_name_declarations) {
    # it was declared within visible scope, so it must be a variable:
    prepend($node,'$');
    append($node,'->');
  };
  
  $node->setNodeName('code');
};

# translate variable names
for my $node ($tree->findnodes('//Identifier[@name = "FunctionName"]')) {
  $node->setNodeName('code');
};

for my $node ($tree->findnodes('//FunctionDeclaration')) {
  prepend($node, q{sub });

  # fudge our formal parameters into the FunctionBody:
  my ($body) = $node->findnodes('./*/FunctionBody');
  my ($params) = $node->findnodes('./*/FormalParameterList');
  $node->removeChild($params);
  if ($params->findnodes('./*')) {
    prepend( $body, q{my }, $params, qq{ = \@_;\n} );
  };

  $node->setNodeName('code');
};

for my $node ($tree->findnodes('//FunctionExpression')) {
  prepend($node, q{sub });

  # fudge our formal parameters into the FunctionBody:
  my ($body) = $node->findnodes('./*/FunctionBody');
  my ($params) = $node->findnodes('./*/FormalParameterList');
  $node->removeChild($params);
  if ($params->findnodes('./*')) {
    prepend( $body, q{my }, $params, qq{ = \@_;\n} );
  };

  $node->setNodeName('code');
};

# Make each FormalParameterList into a comma separated list:
for my $node ($tree->findnodes('//FormalParameterList')) {
  prepend($node, q{(});
  append($node, q{)});

  # fudge our formal parameters into the FunctionBody:
  my @names = $node->findnodes('./Identifier');
  pop @names;
  $node->insertAfter( textnode(','), $_ ) for @names;
  $node->setNodeName('code');
};

# Translate member calls
for my $node ($tree->findnodes('//MemberMemberExpression')) {
  my ($lhs) = $node->findnodes('./*[@name="lhs"]');
  my ($method) = $node->findnodes('./Identifier[@name="member"]');
  my $methodcall = $tree->createTextNode(sprintf q{->});
  $node->insertAfter( $methodcall, $lhs );
  $method->setNodeName( 'code' );
  $node->setNodeName('code');
};

for my $node ($tree->findnodes('//Identifier[@name != "Callee" or not(@name)]')) {
  my ($name) = $node->findnodes('./text()');
  my $newname = $tree->createTextNode(sprintf q{$%s}, $name->textContent);
  $node->replaceChild($newname,$name);
  $node->setNodeName('code');
};

# Translate an initializer:
for my $node ($tree->findnodes('//VariableDeclaration/Initializer')) {
  prepend($node, q{ = });
  $node->setNodeName('code');
};

# Translate a plain variable declaration:
for my $node ($tree->findnodes('//VariableDeclaration')) {
  prepend( $node, q{my });
  append( $node, qq{;\n});
  $node->setNodeName('code');
};

# Function / method call
for my $node ($tree->findnodes('//ArgumentList')) {
  prepend( $node, q{(});
  append( $node, q{)});

  my @names = $node->findnodes('./*');
  pop @names;
  $node->insertAfter( textnode(','), $_ ) for @names;

  $node->setNodeName('code');
};

# Function body
for my $node ($tree->findnodes('//FunctionBody')) {
  prepend( $node, qq({\n));
  append( $node, qq(}\n));
  $node->setNodeName('code');
};

# Function / method call
for my $node ($tree->findnodes('//Expression')) {
  prepend( $node, q{(});
  append( $node, q{)});
  $node->setNodeName('code');
};

# ExpressionStatements should end with a semicolon
for my $node ($tree->findnodes('//ExpressionStatement')) {
  append( $node, qq{;\n});
  $node->setNodeName('code');
};

# ReturnStatements should end with a semicolon
for my $node ($tree->findnodes('//ReturnStatement')) {
  prepend( $node, q{return});
  append( $node, qq{;\n});
  $node->setNodeName('code');
};

eliminate_nodes('//Program');
eliminate_nodes('//SourceElements');
eliminate_nodes('//Statement');
eliminate_nodes('//VariableStatement');
eliminate_nodes('//Number');
eliminate_nodes('//VariableDeclaration');
eliminate_nodes('//VariableDeclarationList');
eliminate_nodes('//AdditiveOperator');
eliminate_nodes('//code');

print "Node left: ",$_->toString,"\n"
  for ($tree->findnodes('/AST//.'));
my $output = q{
  use strict;
{ package perl;
use strict;
  sub print {
    shift;
    CORE::print @_,"\n";
  };
};
my $perl = bless sub {}, "perl";
};

$output .= $_->toString() for ($tree->findnodes('//text()'));
$output =~ s/&gt;/>/g;
print $output;
print "---\n";
eval $output;
print $@ if $@;
