package Javascript::PurePerl;

use strict;
use XML::LibXML;
use Carp qw(croak cluck);
use Data::Dumper;
use Javascript::Grammar;
use base 'Class::Accessor';

use vars qw($VERSION);
$VERSION = '0.09';

__PACKAGE__->mk_accessors(qw( tree yyparser ));

sub new {
  my ($class,%args) = @_;

  my $self = {};
  bless $self,$class;

  my $parser = Javascript::Grammar->new();
  $parser->YYData->{code} = $args{code}
    if exists $args{code};
  $parser->YYData->{pos} = 0;
  $self->yyparser($parser);
  $parser->YYData->{code} = $args{code};

  $self;
};

sub parse {
  my ($self, %args) = @_;
  $self->yyparser->YYData->{code} = $args{code}
    if exists $args{code};
  delete $args{code};
  my $tree = $self->yyparser->YYParse(
                  yylex => \&Javascript::Grammar::yylex,
                  yyerror => sub { $self->error },
		  %args,
                  #yydebug => 0x1F,
             );

  # now transform the tree to an XML::LibXML document:
  $tree = $self->make_document( $tree );
  $self->tree($tree);

  $tree;
};

sub error {
  my $self = shift;
  my $parser = $self->yyparser;
  my $code = \($parser->YYData->{code});
  my $pos = pos $$code;

  my $lines = substr $$code, 0, $pos;
  my $lineno = $lines =~ m!\r?\n!gsm +1;

  die sprintf join( "\n",
                   "Parse error on line %s:",
                   "Got      '%s' ('%s')",
                   "Expected '%s'",
                   "At '%s'"), $lineno,$parser->YYCurtok, $parser->YYCurval,join( " or ", $parser->YYExpect ), substr $$code, $pos;
};

sub optimize {
  my $self = shift;
};

sub append_subtree {
  my ($self,$document,$parent,$tree,$name) = @_;

  croak "Need a document" unless $document;
  croak "Need a parent element" unless $parent;
  croak "Need a tree" unless $tree;

  my ($node,$element_name);
  $element_name = ref $tree;
  cluck "Element without name found"
    unless $element_name;
  $node = $document->createElement( $element_name );
  $node->setAttribute('name',$name)
    if ($name);

  while (@$tree) {
    #warn Dumper $tree;
    my ($childname,$child) = splice @$tree,0,2;

    die "No name found" unless $childname;

    if ($childname eq '__VALUE__') {
      croak sprintf "A __VALUE__ entry must be a string/scalar, %s is a reference (%s)", Dumper($child), ref $child
        if (ref $child and ref $child ne 'Regexp');

      $node->appendChild( $document->createTextNode( $child ))
    } elsif ($childname eq '__ELEMENTS__') {
      croak "An __ELEMENTS__ entry must be an array reference"
        unless (ref $child eq 'ARRAY');

      for (@$child) {
        if (! defined $_ or ! ref $_) {
          warn ref $tree;
          warn Dumper $tree;
        };
        $self->append_subtree($document,$node,$_)
      };
    } elsif ($childname =~ /^__/){
      croak "Unknown special value '$name' found!";
    } else {
        if (! defined $child) {
          warn $childname;
          warn ref $tree;
          warn Dumper $tree;
        };
        $self->append_subtree($document,$node,$child,$childname);
      };
  };
  $parent->appendChild($node);
};

sub make_document {
  my ($self,$tree) = @_;

  my $document = XML::LibXML::Document->createDocument(1,'');
  $document->setStandalone(1);
  my $root = $document->createElement('AST');
  $document->setDocumentElement($root);
  $self->append_subtree($document,$root,$tree);

  return $document;
};

1;
__END__

=head1 NAME

Javascript::PurePerl - Javascript Parser

=head1 SYNOPSIS

=for example begin

  use strict;
  use Javascript::PurePerl;

  my $js = q{alert('Hello World');};

  my $tree = Javascript::PurePerl->new->parse($js);

=for example end

=head1 DESCRIPTION

Simplicistic parser for Javascript

The Javascript gets parsed and returned as an XML
document using L<XML::LibXML>. This makes tree transformations
and searching through the tree much easier. The returned
nodes will have an attribute C<name>, which contains the
name of the subnode. See the included tests for examples
on how to operate on the tree.

No interpretation or compilation is done on the tree yet, not
even constant folding and the like.

=head2 EXPORT

None by default.

=head1 SEE ALSO

L<XML::LibXML>,L<Parse::RecDescent>, L<Javascript>

=head1 AUTHOR

Max Maischein E<lt>corion@cpan.orgE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2003,2004 by Max Maischein

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.2 or,
at your option, any later version of Perl 5 you may have available.

=cut
